// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   Grayscale Image
 *
 * The present package implements all the basic algorithms dealing
 * with grayscale images and their rectangular parts. Particular
 * attention is given to image arithmetics and image streams.
 *
 * An image is a pixel map, a matrix of pixels arranged row-by-row.
 * Each pixel specifies the gray level at a particular point. A row or
 * a column pixel index always starts with 0.
 *
 * The present package provides all facilities to completely AVOID
 * returning images. If you really needs to return an image, you
 * should return a LazyImage object instead. The conversion is
 * completely transparent to the end user, e.g. "IMAGE im1 =
 * expand_twice(im2);". It is also as efficient as possible.
 *
 * Please see README for more details and explanations.
 *
 * $Id: image.h,v 3.0 2000/12/05 01:14:21 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifndef __GNUC__
#pragma once
#endif
#ifndef _image_h
#define _image_h

#if defined(__GNUC__)
#pragma interface
#endif

#include "myenv.h"
#include "std.h"

typedef unsigned short GRAY;		// Pixel type
typedef signed short GRAY_SIGNED;	// Pixel type, signed
typedef unsigned int card;		// cf. CARDINAL in Modula-2
const card GRAY_MAXBIT = 8*sizeof(GRAY);	// Max no of bits per pixel
const card GRAY_MAXVAL = ((1<<GRAY_MAXBIT)-1);

/*
 *------------------------------------------------------------------------
 *			Auxiliary classes
 */

class Extrema;
class IMAGE;

				// An interface describing an operation to 
				// apply to a pixel.
				// This interface is used by IMAGE::apply()
				// and similar for-each-type iterators. They
				// call a concrete instance of this interface
				// on each element of an image etc. collection
				// under consideration.
				// The operation receives the (i,j) index of
				// the current pixel and its _value_:
				// thus this interface may not modify the
				// element itself.
class ConstPixelAction
{
public:
  virtual void operation(const GRAY pixel, const card row, const card col) = 0;
};

				// The same as above, only an operation
				// is allowed to modify the pixel
				// it is being applied to.
class PixelAction
{
public:
  virtual void operation(GRAY& pixel, const card row, const card col) = 0;
};


				// A Lazy image constructor
				// -- a promise of an image.
				// If you're writing a function that
				// is to build and return an image
				// (e.g., an image reader), don't make
				// this function return an IMAGE
				// object or an IMAGE*. Return a
				// LazyImage object instead.
struct LazyImage
{
  friend class IMAGE;
  virtual void fill_in(IMAGE& im) const = 0;
protected:
  card nrows, ncols, depth;
public:
  LazyImage(const card _nrows, const card _ncols, const card _depth)
    : nrows(_nrows), ncols(_ncols), depth(_depth) {}
  virtual ~LazyImage(void) {}
};

			// The following class is a protecting
			// envelope for an abstract LazyImage
			// pointer. The envelope passes from
			// read_pgm(), read_tiff() and similar
			// functions to an IMAGE constructor. When the
			// sender places a ptr to a concrete LazyImage
			// object into the envelope, the sender
			// thereby disclaims the ownership of the
			// object. The object will be disposed of by
			// the corresponding IMAGE constructor. The
			// purpose of introducing the extra
			// indirection is to keep classes internal to
			// TIFF, PGM, XWD subsystems from the general
			// image.h header.

class LazyImagePtr
{
  friend class IMAGE;
  LazyImage * lip;
public:
  LazyImagePtr(LazyImage * _lip) : lip(_lip) {}
};

/*
 *------------------------------------------------------------------------
 *		Position specification and operations on it
 */

class rowcol {			// Specifying the row/col position
  friend class Extrema;

protected:
  card row_val;			// both row, col start from 0
  card col_val;

public:
					// Constructors
  rowcol(const int row, const int col) 	: row_val(row), col_val(col) {}
//??  rowcol(void)				: row_val(-1), col_val(-1) {}

  card row(void) const			{ return row_val; }
  card col(void) const			{ return col_val; }

  bool operator == (const rowcol& pos) const
  { return row_val == pos.row_val && col_val == pos.col_val; }

					// Offset the current position
  rowcol& operator += (const rowcol& pos)
  	{ row_val += pos.row_val; col_val += pos.col_val; return *this; }
  rowcol& operator -= (const rowcol& pos)
  	{ row_val -= pos.row_val; col_val -= pos.col_val; return *this; }

  friend inline rowcol operator + (const rowcol& pos1, const rowcol& pos2);
  friend inline rowcol operator - (const rowcol& pos1, const rowcol& pos2);

					// Scale the current position
  rowcol& operator *= (const int scalef)
  	{ row_val *= scalef; col_val *= scalef; return *this; }

  friend inline rowcol operator *  (const rowcol& pos, const int scalef);
  friend inline rowcol operator << (const rowcol& pos, const int shiftf);
  friend inline rowcol operator >> (const rowcol& pos, const int shiftf);
};

inline rowcol operator + (const rowcol& pos1, const rowcol& pos2)
	{ return rowcol(pos1.row_val+pos2.row_val, pos1.col_val+pos2.col_val);}
inline rowcol operator - (const rowcol& pos1, const rowcol& pos2)
	{ return rowcol(pos1.row_val-pos2.row_val, pos1.col_val-pos2.col_val);}

inline rowcol operator *  (const rowcol& pos, const int scalef)
	{ return rowcol(pos.row_val*scalef, pos.col_val*scalef); }
inline rowcol operator >> (const rowcol& pos, const int shiftf)
	{ return rowcol(pos.row_val >> shiftf, pos.col_val >> shiftf); }
inline rowcol operator << (const rowcol& pos, const int shiftf)
	{ return rowcol(pos.row_val << shiftf, pos.col_val << shiftf); }

				// This is only a specification of a rectangle
				// area
class rectarea
{
  rowcol upperleft;
  rowcol lowright;
public:
				// Specify the rect area by the coordinates of
				// the upper-left and lower-right corners
  rectarea(const rowcol _ul, const rowcol _lr) : 
	upperleft(_ul), lowright(_lr) {}
  
};

class Extrema			// Find min/max values of the (signed) image
{
  GRAY_SIGNED max_value;		// Max pixel value in the image
  GRAY_SIGNED min_value;		// Min pixel value in the image

  rowcol      max_pixel;		// Position of the largest pixel
  rowcol      min_pixel;		// Position of the smallest pixel

public:
  Extrema(const IMAGE& image);		// Find extreme pixels of the image

  GRAY_SIGNED max(void) const		{ return max_value; }
  GRAY_SIGNED min(void) const		{ return min_value; }

  const rowcol& loc_max(void) const		{ return max_pixel; }
  const rowcol& loc_min(void) const		{ return min_pixel; }
};

/*
 *------------------------------------------------------------------------
 *	Special kinds of images that have to be forward-declared
 */
class Rectangle;
class FilterIt;
class TIFFUserAction;


				// Lazy Expansion
				// Expand the prototype twice in each
				// dimension
class LazyExpandedImage : public LazyImage
{
  const IMAGE& proto;
  void fill_in(IMAGE& m) const;
  inline LazyExpandedImage(const IMAGE & _proto);
public:
  friend inline const LazyExpandedImage expand_twice(const IMAGE & proto);
};

				// Lazy Shrinking, twice in each dimension
class LazyShrunkImage : public LazyImage
{
  const IMAGE& proto;
  void fill_in(IMAGE& m) const;
  inline LazyShrunkImage(const IMAGE & _proto);
public:
  friend const LazyShrunkImage shrink_twice(const IMAGE & proto);
};

/*
 *------------------------------------------------------------------------
 *			The Image class itself
 *
 */

class IMAGE
{
  friend class Rectangle;
  friend class Extrema;
  friend class FilterIt;
  friend class Image_istream;
  friend class Image_iostream;
  friend class PixelWiseConst;

private:			// Private part
  int valid_code;			// Validation code
  enum { IMAGE_val_code = 577767 };	// Image validation code
  card ncols;				// Image width in pixels
  card nrows;				// Image height in pixels
  unsigned long int npixels;		// Total no of pixels in the image
  const char * name;			// Image name
  card bits_per_pixel;			// Image depth
  GRAY ** scanrows;			// scanrows[i] = &pixels[i,0]
  GRAY * pixels;			// pixels[i,j] is the
  					// pixel value at point (i,j)
					// Pixels are ordered in rows

  void allocate(const card nrows, const card ncols, const card depth);

  friend void LazyExpandedImage::fill_in(IMAGE& m) const;
  friend void LazyShrunkImage::fill_in(IMAGE& m) const;


public:			// Public interface
  
				// Constructors and destructors
  					// Make a blank image
  IMAGE(const card nrows, const card ncols, const card depth);
  IMAGE(const IMAGE& image);		// Copy constructor: expensive
  IMAGE(const Rectangle& area);		// Make an image from a rect area
					// of another image

  IMAGE(const LazyImage& lazy_constructor);//Make an image using given recipe
  IMAGE(const LazyImagePtr& lazy_constructor);

  ~IMAGE(void);

  volatile void invalid_image(void) const; // The image in question is not 
  					// valid ==> crash the program
  void is_valid(void) const
  { if(valid_code != IMAGE_val_code) invalid_image(); }

				// Status inquiries
  card q_nrows() const			{ return nrows; }
  card q_ncols() const			{ return ncols; }
  card q_depth() const			{ return bits_per_pixel; }
  unsigned long int q_npixels() const	{ return npixels; }
  const char * q_name() const  		{ return name; } 
  void set_name(const char * new_name);	

				// Individual pixel manipulations
  const GRAY operator () (const card row, const card col) const;
  GRAY& operator () (const card row, const card col);
  const GRAY operator () (const rowcol pos) const;
  GRAY& operator () (const rowcol pos);
  void  sure_within(const rowcol pos) const;	// Make sure the pos


				// Single image operations
  IMAGE& clear(void);			// Clear the image
  IMAGE& invert(void);			// Invert the image

				// Apply a user-defined action
  IMAGE& apply(const PixelAction& action);	// to each pixel
  const IMAGE& apply(const ConstPixelAction& action) const;


				// Two images operations
  					// Although the following method does
  					// not need a privileged access to an
  					// IMAGE, it is required to be a member
  					// of the class
  IMAGE& operator = (const GRAY val);	// Assignment
  IMAGE& IMAGE::operator = (const IMAGE& source);
					// Assign another to '*this' resizing
  IMAGE& coerce(const IMAGE& another);	// '*this' as necessary to fit (by
					// not necessarily int factor!)

  IMAGE& operator = (const LazyImage& source);// Force the promise

  bool operator == (const IMAGE& im2) const;
  friend void compare(const IMAGE& im1, const IMAGE& im2, 
		      const char * title);
  friend inline void are_compatible(const IMAGE& im1, const IMAGE& im2);

				// I/O: write, read, display, print info
  					// Write to a file
					// "| command name" is OK as a file
					// name
  void write(const char * file_name,const char * title = "") const
		{ write_xwd(file_name,title); }
					// Write an X window dump
  void write_xwd(const char * file_name,const char * title = "") const;
					// Write a Portable GrayMap
  void write_pgm(const char * file_name,const char * title = "") const;
					// Write a TIFF file
  void write_tiff(const char * file_name,const char * title = "") const;
  void write_tiff(const char * file_name,const char * title,
		  const TIFFUserAction& user_adding_tags) const;
  void display(const char * title) const;
  void info(void) const;		// Print the info about the image
  void print(const char * title) const;	// Print the image as a table

				// Clip a square area of the image
  Rectangle	square_of (const card size, const rowcol pos);
				// Clip a rectangular area of the image
				// specified by the coordinates of
				// the upper-left and lower-right corners
  Rectangle	rectangle (const rowcol uppleft, const rowcol lowright);

};

				// Estimate the norm of the (signed) image
double norm_1(const IMAGE&);		// SUM{ |(signed)pixel[i,j]| }
double norm_2_sqr(const IMAGE&);	// SUM{ (signed)pixel[i,j]^2 }
int    norm_inf(const IMAGE&);		// MAX{ |(signed)pixel[i,j]| }

				// Estimate the norm of the difference 
				// between two (signed) image
					// SUM{ |(signed)pixel[i,j]| }
double norm_1(const IMAGE& im1, const IMAGE& im2);
					// SUM{ (signed)pixel[i,j]^2 }
double norm_2_sqr(const IMAGE& im1, const IMAGE& im2);
					// MAX{ |(signed)pixel[i,j]| }
int    norm_inf(const IMAGE& im1, const IMAGE& im2);

			// Reading of an image from a file or a stream, in
			// a particular file format.
const LazyImagePtr			// read a Portable GrayMap
read_pgm(const char file_name[], const bool verbose);
const LazyImagePtr			// read an X Window dump
read_xwd(const char file_name[], const bool verbose);
const LazyImagePtr			// read TIFF
read_tiff(const char file_name[], const bool verbose);

					// Read an image from the file,
					// attempting to detect the file format
const LazyImagePtr
read_image(const char file_name[], const bool print_header_info = false);

/*
 * Functions/procedures (generally grouped together under a ElementWise
 * class umbrella) that perform a variety of operation on each element
 * of some structure in turn.
 *
 * PixelWiseConst and PixelWise classes are largely a syntactic
 * sugar; you can't explicitly construct an object of this class, on stack
 * or on heap. The only way the class can be used is within a pattern
 *      to_every(image) = 1;
 *	if( of_every(image) > 1 ) ...
 * etc. Besides, this is the only way it makes sense.
 *
 * Note that the PixelWise classes are friends of IMAGE, so they are
 * privy of IMAGE internals and can use them to make processing faster.
 */


				// A functor describing an operation to apply 
				// to every element of a collection in
				// turn
				// This operation receives only the elements's
				// value, and thus it may not modify the
				// element under consideration.
class PixelWiseConstAction
{
				// Those aren't implemented; but making them
				// private forbids the assignment
  PixelWiseConstAction(const PixelWiseConstAction&);
  PixelWiseConstAction& operator= (const PixelWiseConstAction&);
public:
  PixelWiseConstAction(void) {}
  virtual void operator () (const GRAY pixel) = 0;
};

				// The same as above but this functor is given
				// a reference to an element, which it can
				// modify
class PixelWiseAction
{
				// Those aren't implemented; but making them
				// private forbids the assignment
  PixelWiseAction(const PixelWiseAction&);
  PixelWiseAction& operator= (const PixelWiseAction&);
public:
  PixelWiseAction(void) {}
  virtual void operator () (GRAY& pixel) = 0;
};

class PixelWiseConst
{
  friend class PixelWise;

  GRAY * const start_ptr;		// Pointer to the beginning of
  					// the group of pixels
  GRAY * const end_ptr;			// Points after the end of the group
  const card bits_per_pixel;		// Image depth
  
  void operator=(const PixelWiseConst&);// is not implemented, ergo, is
                                        // not allowed
  PixelWiseConst(const PixelWiseConst&);// Cloning is not allowed, either
  
				// A private constructor, to make
				// sure the object can't be constructed
				// and left, while the image would disappear
				
				// Use of_every() friend to do
				// the actual construction
  PixelWiseConst(const IMAGE& im)
  	: start_ptr(const_cast<GRAY*>(im.pixels)),
  	  end_ptr(const_cast<GRAY*>(im.pixels+im.npixels)),
	  bits_per_pixel(im.bits_per_pixel)
  	{ im.is_valid(); }

  //  inline PixelWiseConst(const ConstImageRow& imr);
  
  				// Just a helper class to disambiguate 
  				// type conversion from PixelWiseConst
  				// to PixelWiseStrideConst and others
  struct PixelWiseConstRef
  {
    const PixelWiseConst& ref;
    PixelWiseConstRef(const PixelWiseConstRef&);
    void operator = (const PixelWiseConstRef&);
    PixelWiseConstRef(const PixelWiseConst& _ref) : ref(_ref) {}
  };

public:

					// No public constructors...

                                        // PixelWiseConst(IMAGE& im) would be
                                        // implicitly called
  friend inline PixelWiseConst of_every(const IMAGE& im) { return im; }
  //friend inline PixelWiseConst of_every(const IMAGE::ConstReference& im)
  //	{ return im.ref(); }
  
  //friend inline PixelWiseConst of_every(const ConstImageRow& imr);
 				
  				// Every PixelWiseConst is a
  				// PixelWiseStrideConst with a stride of 1
  //  inline operator PixelWiseStrideConst (void);
  
				// Comparisons
				// Find out if the predicate
				// "(signed)pixel op val" is true for ALL
				// pixels of the group?
  bool operator ==  (const GRAY val) const;	// ? all pixels == val
  bool operator !=  (const int val) const;	// ? all (signed)pixels != val
  bool operator <   (const int val) const;	// ? all (signed)pixels <  val
  bool operator <=  (const int val) const;	// ? all (signed)pixels <= val
  bool operator >   (const int val) const;	// ? all (signed)pixels >  val
  bool operator >=  (const int val) const;	// ? all (signed)pixels >= val
  
				// Find out if the predicate
				// "(signed)pixel op another.pixel" is true
				// for ALL pixels of the two groups
  bool	  operator ==  (const PixelWiseConst& another) const;
  bool	  operator !=  (const PixelWiseConst& another) const;
  bool	  operator <   (const PixelWiseConst& another) const;
  bool	  operator <=  (const PixelWiseConst& another) const;
  bool	  operator >   (const PixelWiseConst& another) const;
  bool	  operator >=  (const PixelWiseConst& another) const;
				
  void sure_compatible_with(const PixelWiseConst& another) const
  	{ assure(end_ptr-start_ptr == another.end_ptr-another.start_ptr,
  		"Two groups of pixels to operate on have different sizes"); }
  
  				// Data reduction: reduce a collection
  				// to one value: a "norm" of the collection
  double sum(void) const;
  double sum_squares(void) const;
  double sum_abs(void) const;
  int max_abs(void) const;
  //double foldr(const double seed, Reductor& reductor);
  
  				// Reduce a difference between two collections
  				// to a single value: a "norm" of the
  				// difference
  double sum_squares(const PixelWiseConst& another) const;
  double sum_abs(const PixelWiseConst& another) const;
  int max_abs(const PixelWiseConst& another) const;
						// Scalar product
  double operator * (const PixelWiseConst& another) const;
  //double foldr(const double seed, Reductor2& reductor);
  
  				// Just let the user do what he wants
  				// with each element in turn
  				// (without modifying it, of course)
  PixelWiseConstAction& apply(const PixelWiseConstAction& functor) const;
};


class PixelWise : public PixelWiseConst
{
  void operator=(const PixelWise&);	// is not implemented, ergo, is
                                        // not allowed
  PixelWise(const PixelWise&);	// Cloning is not allowed, either
  
				// A private constructor, to make
				// sure the object can't be constructed
				// and left, while the image would disappear
				
				// Use a to_every() friend to do
				// the actual construction
  PixelWise(IMAGE& im) : PixelWiseConst(im) {}
  //inline PixelWise(const ImageRow& imr);
public:

					// No public constructors...

                                        // PixelWise(IMAGE& im) would be
                                        // implicitly called
  friend inline PixelWise to_every(IMAGE& im) { return im; }
  //  friend inline PixelWise to_every(const IMAGE::Reference& im)
  //  	{ return im.ref(); }
//  friend inline PixelWiseConst of_every(const ConstMatrixColumn& mc);
  // friend inline PixelWise to_every(const ImageRow& imr); // see below

   				// Every PixelWise is a
  				// PixelWiseStride with a stride of 1
  // inline operator PixelWiseStride (void);
  				// The same as the conversion operator above,
  				// use when a compiler (gcc 2.8.1) fails to
  				// apply it
  // inline PixelWiseStride with_stride (void);

				// group-scalar arithmetics
				// Modify every pixel of the group
				// according to the operation
  void operator =   (const GRAY val);	// Assignment to all the elems
  void operator -=  (const int val);		// Diminish the brightness
  void operator +=  (const int val);		// Increase the brightness
  void operator *=  (const int val);
  void operator |=  (const int val);		// OR
  void operator &=  (const int val);		// AND
  void operator ^=  (const int val);		// XOR
  void operator <<= (const card val);		// Shift all the pixels
  void operator >>= (const card val);		// Shift all the pixels

				// Other element-wise image operations
  void clear(void);			// Clear the collection
  void invert(void);			// Invert the collection
  void abs(void);			// pixel = |(signed)pixel|
  void clip_to_intensity_range(void);	// Clip pixel values to
					// [0,1<<bits_per_pixel-1]
  void normalize_for_display(void);	// Normalize pixel values to be
					// in range 0..1<<bits_per_pixel-1
  void equalize(const int no_grays);	// Perform the histogram equalization

				// Two-group operations
  void operator =  (const PixelWiseConst& another);
  void operator += (const PixelWiseConst& another);
  void operator -= (const PixelWiseConst& another);
  void operator |= (const PixelWiseConst& another);
  void operator &= (const PixelWiseConst& another);
  void operator ^= (const PixelWiseConst & another);

//??  void add(IMAGE& target, const int scalar,const IMAGE& source);
						// Shift the source to 'pos'
						// clip if necessary
						// multiply by scalar
						// and add
//  void shift_clip_add(rowcol pos, const int scalar, const IMAGE& source);


  PixelWiseAction& apply(const PixelWiseAction& functor);
};


/*
 *------------------------------------------------------------------------
 *		 Dealing with the rectangular area of the image
 */

class Rectangle
{
  friend class IMAGE;

  IMAGE&	image;		// The image I'm a rectangle of
  card		nrows;		// Dimension of the rectangle
  card 		ncols;
  GRAY *	ptr;		// Pointer to the upper left corner of 
				// the rectangle
  int		inc_to_nextrow;	// inc_to_nextrow = p' - p =
  				// 		  = image.ncols - ncols
				// where
				// p = ptr + ncols points to the pixel
				// just beyond the current scanline of
				// the Rectangle
				// p' = (ptr - col) + image.ncols + col
				// points to the first pixel of the next
				// scanline of the rectangtle
				// col tells the column of the upper left
				// corner of the rectangle within the image
 
				// Private constructor
  				// Note these are private constructors to
				// be used with IMAGE::square_of/rectangle
				// functions
  Rectangle
    (IMAGE& im, const card size, const rowcol pos);
  Rectangle
    (IMAGE& im, const rowcol uppleft, const rowcol lowright);

  public:
				// Note how to construct square area of
				// the image 'im':
				//	im.square_of(10,rowcol(1,5))
				// and rectangular area
				// 	im.rectangle(rowcol(0,1),rowcol(4,5))

  Rectangle (IMAGE& im);		// Treat entire image as a rectangle
					// Assign a value to all the pixels
					// in the rectangle area
  Rectangle& operator  = (const GRAY val);
  					// Modify the pixels in the rectangle
					// area
  Rectangle& operator +=  (const int val);
  Rectangle& operator -=  (const int val);
  Rectangle& operator *=  (const int val);
  Rectangle& operator |=  (const int val);
  Rectangle& operator &=  (const int val);
  Rectangle& operator ^=  (const int val);
  Rectangle& operator <<= (const card val);
  Rectangle& operator >>= (const card val);

					// Copy a rectangle a_rect into the
					// rectangular area of the image
  Rectangle& operator = (const Rectangle& a_rect);

  					// Get a total sum of all the pixels
  friend double sum_over(const Rectangle& sa);
};

/*
 *------------------------------------------------------------------------
 *	The following functions let one traverse an image _sequentially_
 * 		 pixel-by-pixel in the row major order
 * Indeed, in many situations, random access to pixels is unnecessary.
 * The classes below act as input/output "image" streams. As with regular
 * streams, be sure to check for eof() before attempting to get() to the
 * pixel.
 * Note, since all the methods of the image streams are implemented inline,
 * the whole iteration would be inline, too.
 */

				// The following class is an STL-like iterator
				// to access IMAGE pixel _values_ one-by-one
				// in a row-by-row fashion
				// Note this is a _read-only_ stream
class Image_istream
{
  const GRAY * curr_pixel_p;		// curr_pixel_p <= last_pixel_p always
  const GRAY * const last_pixel_p;	// The = condition means EOF
public:
  Image_istream(const IMAGE& image)
  	: curr_pixel_p(image.pixels),
  	  last_pixel_p(image.pixels+image.npixels)
  	  { image.is_valid(); }
  bool eof(void) const		{ return curr_pixel_p == last_pixel_p; }

  	  			// Get a current pixel and advance in the
  	  			// stream
  GRAY get(void)
  	{ if( curr_pixel_p >= last_pixel_p)
  	   _error("Can't get from Image_istream past the image boundary!");
  	  return *curr_pixel_p++; }
  GRAY peek(void) const		// Does *not* advance the "stream ptr"
  	{ if( curr_pixel_p >= last_pixel_p)
  	   _error("Can't peek from Image_istream past the image boundary!");
  	  return *curr_pixel_p; }
};

				// This class provides read *and* write
				// access to image pixels, one-by-one
class Image_iostream
{
  GRAY * curr_pixel_p;			// curr_pixel_p <= last_pixel_p always
  const GRAY * const last_pixel_p;	// The = condition means EOF
public:
  Image_iostream(IMAGE& image)
  	: curr_pixel_p(image.pixels),
  	  last_pixel_p(image.pixels+image.npixels)
  	  { image.is_valid(); }
  bool eof(void) const		{ return curr_pixel_p == last_pixel_p; }
  	  			// Get a current pixel and advance in the
  	  			// stream
  GRAY& get(void)
  	{ if( curr_pixel_p >= last_pixel_p)
  	   _error("Can't get from Image_istream past the image boundary!");
  	  return *curr_pixel_p++; }
  GRAY& peek(void) const	// Does *not* advance the ptr
  	{ if( curr_pixel_p >= last_pixel_p)
  	   _error("Can't peek from Image_istream past the image boundary!");
  	  return *curr_pixel_p; }
};

/*
 *------------------------------------------------------------------------
 *			Some service procedures
 */

inline
int log2(			// Find a binary logarithm of n
	const int n )		// and check that n is a power of two
{
  register int i,k;

  assure(n > 0, "log2: the argument's got to be positive!");

  for(n & 0xff ? k=0,i=1 : k=8,i=256; i < n; k++, i*=2)
    ;
  if( i != n )
    _error("log2: the argument %d has got to be an exact power of two",n);

  return k;
}

inline
int exp2(const int k)			// Compute 2^k, k>=0
{
  assure(k >= 0, "exp2: the argument may not be negative!");
  if( k > (signed)sizeof(int)*8-1 )
    _error("exp2: the exponent %d is too big",k);

  return 1<<k;
}

inline					// return b * round( a/b )
int round_to_even_multiple(const int a, const int b)
{
  assert( b > 0 );
  register int rem = a % b;
  if( 2*rem < b )
    return a - rem;
  else
    return a + b - rem;
}

				// Service functions (useful in the
				// verification code). They print some detail
				// info if the validation condition fails
void verify_pixel_value(const IMAGE& im, const GRAY val);

/*
 *------------------------------------------------------------------------
 *			Inline Image Procedures
 */

inline IMAGE::IMAGE(const card no_rows, const card no_cols, const card depth)
{
  allocate(no_rows,no_cols,depth);
}

					// Copy constructor, expensive: use
					// sparingly
inline IMAGE::IMAGE(const IMAGE& old)
{
  old.is_valid();
  allocate(old.nrows,old.ncols,old.bits_per_pixel);
  *this = old;
}


inline IMAGE::IMAGE(const LazyImage& lazy_constructor)
{
  allocate(lazy_constructor.nrows,lazy_constructor.ncols,
	   lazy_constructor.depth);
  lazy_constructor.fill_in(*this);
}

inline IMAGE::IMAGE(const LazyImagePtr& lazy_cptr)
{
  allocate(lazy_cptr.lip->nrows,lazy_cptr.lip->ncols,
	   lazy_cptr.lip->depth);
  lazy_cptr.lip->fill_in(*this);
  delete lazy_cptr.lip;
  const_cast<LazyImagePtr&>(lazy_cptr).lip = 0;
}

				// Force a promise of an image
				// That is, apply a lazy_constructor
				// to the current image
inline IMAGE& IMAGE::operator = (const LazyImage& lazy_constructor)
{
  is_valid();
  if( nrows != lazy_constructor.nrows || ncols != lazy_constructor.ncols
      || bits_per_pixel != lazy_constructor.depth )
    info(),
    _error("The image above is incompatible with the assigned "
    	   "Lazy image");
      
  lazy_constructor.fill_in(*this);
  return *this;
}

inline GRAY& IMAGE::operator () (const card row, const card col)
{
  is_valid();
  if( row >= nrows )
    _error("Row index %d is out of image boundaries [0,%d]",row,nrows-1);
  if( col >= ncols )
    _error("Col index %d is out of image boundaries [0,%d]",col,ncols-1);
  
  return (scanrows[row])[col];
}

inline const GRAY IMAGE::operator () (const card row, const card col) const
{ return (*(const_cast<IMAGE*>(this)))(row,col); }

inline const GRAY IMAGE::operator () (const rowcol pos) const
{ return operator()(pos.row(),pos.col()); }

inline GRAY& IMAGE::operator () (const rowcol pos)
{ return operator()(pos.row(),pos.col()); }

					// Make sure the pos is within the
					// image
inline void  IMAGE::sure_within(const rowcol pos) const
{
  is_valid();
  if( pos.row() >= nrows )
    _error("Row index %d is out of image boundaries [0,%d]",pos.row(),nrows-1);
  if( pos.col() >= ncols )
    _error("Col index %d is out of image boundaries [0,%d]",pos.col(),ncols-1);
}

inline IMAGE& IMAGE::clear(void)	// Clean the image
{
  is_valid();
  memset(pixels,0,npixels*sizeof(GRAY));
  return *this;
}


inline void are_compatible(const IMAGE& im1, const IMAGE& im2)
{
  im1.is_valid();
  im2.is_valid();
  
  if( im1.ncols != im2.ncols || im1.nrows != im2.nrows )
    _error("The image %dx%d and the image %dx%d have different sizes",
	   im1.nrows,im1.ncols,im2.nrows,im2.ncols);
}

				// Apply a user-defined action to each
				// pixel of a collection
inline PixelWiseConstAction&
PixelWiseConst::apply(const PixelWiseConstAction& functor) const
{
  register const GRAY * cp = start_ptr;
  while( cp < end_ptr )
    const_cast<PixelWiseConstAction&>(functor)(*cp++);
  return const_cast<PixelWiseConstAction&>(functor);
}

inline PixelWiseAction&
PixelWise::apply(const PixelWiseAction& functor)
{
  register GRAY * cp = start_ptr;
  while( cp < end_ptr )
    const_cast<PixelWiseAction&>(functor)(*cp++);
  return const_cast<PixelWiseAction&>(functor);
}

				// Construct a square rectangle
inline Rectangle::Rectangle
    (IMAGE& im, const card sq_size, const rowcol pos)
    : image(im), nrows(sq_size), ncols(sq_size)
{
  image.is_valid();

  if( pos.row() >= image.nrows ||
      pos.row() + nrows > image.nrows ||
      pos.col() >= image.ncols ||
      pos.col() + ncols > image.ncols )
    _error("Square area (left upper point [%d,%d], size %d)\n"
	   "is not within the image %dx%d",
	   pos.row(),pos.col(),sq_size,image.nrows,image.ncols);

  ptr  = &(image.scanrows[pos.row()][pos.col()]);
  inc_to_nextrow = image.ncols - ncols;
}

				// Clip a square area from the image
inline Rectangle IMAGE::square_of(const card size, const rowcol pos)
{ return Rectangle(*this, size, pos); }


				// Make a rectangle with upper left corner
				// at uppleft position and lower right corner
				// at lowright position
inline Rectangle::Rectangle
    (IMAGE& im, const rowcol uppleft, const rowcol lowright)
    : image(im), nrows(lowright.row()-uppleft.row()+1), 
                 ncols(lowright.col()-uppleft.col()+1)
{
  image.is_valid();

  image.sure_within(uppleft);
  image.sure_within(lowright);

  ptr  = &(image.scanrows[uppleft.row()][uppleft.col()]);
  inc_to_nextrow = image.ncols - ncols;
}


				// Clip a rectangular area from the image
inline Rectangle IMAGE::rectangle(const rowcol uppleft,const rowcol lowright)
{ return Rectangle(*this, uppleft, lowright); }


				// Treat the entire image as a rectangular area
inline Rectangle::Rectangle (IMAGE& im)
    : image(im), nrows(im.nrows), ncols(im.ncols)
{
  image.is_valid();

  ptr  = image.pixels;
  inc_to_nextrow = 0;
}

				// Make a zero image
class LazyZeroImage : public LazyImage
{
  void fill_in(IMAGE&) const {}
  LazyZeroImage(const IMAGE& proto)
    : LazyImage(proto.q_nrows(),proto.q_ncols(),proto.q_depth()) {}
public:
  friend const LazyZeroImage zero(const IMAGE& proto)
    { proto.is_valid(); return proto; }
};

				// Make an image twice the size of the
				// proto image
inline LazyExpandedImage::LazyExpandedImage(const IMAGE & _proto)
  	: LazyImage(2*_proto.q_nrows(),2*_proto.q_ncols(),
		    _proto.q_depth()),
  	  proto(_proto) {}
inline const LazyExpandedImage expand_twice(const IMAGE & proto)
  	{ proto.is_valid(); return proto; }

				// Make an image half the size of the
				// proto image
				// The image row and column
				// dimensions are assumed to be even numbers
LazyShrunkImage::LazyShrunkImage(const IMAGE & _proto)
  	: LazyImage(_proto.q_nrows()/2,_proto.q_ncols()/2,
		    _proto.q_depth()),
  	  proto(_proto) {}

				// The following is just a syntactic sugar
				// to avoid typing of_every and to_every
inline IMAGE& IMAGE::operator = (const GRAY val)
	{ to_every(*this) = val; return *this; }

inline bool operator ==  (const IMAGE& im, const GRAY val)
	{ return of_every(im) == val; }


#endif
