// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   Grayscale Image
 *
 *	           Read an image from the X Window dump file
 *
 * $Id: read_xwd.cc,v 3.0 2000/10/20 02:32:53 oleg Exp oleg $
 *
 ************************************************************************
 */

#include "image.h"
#include "x11wd.h"
#include "endian_io.h"
#include <iostream.h>

/*
 *------------------------------------------------------------------------
 *			Create an image from the file
 * This function attempts to determine image file format by looking at
 * the first two bytes of the file. They are
 *	'Px' - for a Portable GrayMap
 *	'MM' or 'II' - for a TIFF file
 *	otherwise, it _might_ be an X window dump.
 * After the file format is guessed, the program calls an appropriate
 * reader to actually load the file.
 *
 * Note, we could've passed the EndianIn stream we've opened to
 * the specific reader. In that case we should pass the ptr to a new
 * EndianIn stream rather than a reference, because in the scenario
 * below the EndianIn stream will be closed when read_image() exits.
 * Passing pointers is a mess that is better avoided (e.g., this raises
 * the whole tangle of ownership). Really C++ would've benefitted
 * from a GC.
 */

const LazyImagePtr
read_image(const char file_name[], const bool print_header_info)
{
  cerr << "\nReading the file '" << file_name << "'" << endl;
  EndianIn file(file_name);
  assert( file.good() );

  char byte1 = file.peek();
  if( byte1 == 'P' )
    return read_pgm(file_name,print_header_info);
  else if( byte1 == 'M' || byte1 == 'I' )
    return read_tiff(file_name,print_header_info);
  else
    return read_xwd(file_name,print_header_info);
}

/*
 *------------------------------------------------------------------------
 *			   Class XWDump
 *	 	designed to contain image description
 *		as defined in the X Window Dump file 
 */

class XWDump			// All the info pertaining to the XWD file
  :  public LazyImage, public X11WDFileHeader
{
  char * win_name;		      	// Name of the image in the file
public:
  mutable EndianIn ins;
  XWDump(const char file_name[]);		// Constructor
  void info(void) const;			// Dump the header information
  void fill_in(IMAGE& im) const;
  ~XWDump(void) { if(win_name) delete win_name; }
};

/*
 *------------------------------------------------------------------------
 *			Construct the XWDump
 *	   by reading the header of the X Window dump file
 */

XWDump::XWDump(const char file_name[]) : LazyImage(0,0,0), win_name(0),
					 ins(file_name)
					 
{
  assert(ins.good());
  streampos beg_of_file = ins.tellg();	// Remember the curr position
  ins.set_bigendian();		// Try reading a magic byte in one order
  header_size = ins.read_long("Reading the header_size");
  file_version = ins.read_long("Reading the XWD file version");

  if( file_version != X11WD_FILE_VERSION )
  {
    ins.set_littlendian();		// Try now another byte order
    ins.seekg(beg_of_file);
    header_size = ins.read_long("Reading the header_size");
    file_version = ins.read_long("Reading the XWD file version");
  }

  assure( file_version == X11WD_FILE_VERSION,
	 "I have tried different byte orders but failed to read\nthe XWD "
	 "file version correctly" );
  assure( header_size >= sizeof(X11WDFileHeader),"XWD header is too small" );

  				// Read the rest of the header
  pixmap_format    = ins.read_long("Reading the pixmap_format");
  pixmap_depth     = ins.read_long("Reading the pixmap_depth");
  pixmap_width     = ins.read_long("Reading the pixmap_width");
  pixmap_height    = ins.read_long("Reading the pixmap_height");
  xoffset          = ins.read_long("Reading the xoffset");
  xbyte_order      = ins.read_long("Reading the byte_order");
  bitmap_unit      = ins.read_long("Reading the bitmap_unit");
  bitmap_bit_order = ins.read_long("Reading the bitmap_bit_order");
  bitmap_pad       = ins.read_long("Reading the bitmap_pad");
  bits_per_pixel   = ins.read_long("Reading the bits_per_pixel");
  bytes_per_line   = ins.read_long("Reading the bytes_per_line");
  visual_class     = ins.read_long("Reading the visual_class");
  red_mask         = ins.read_long("Reading the red_mask");
  green_mask       = ins.read_long("Reading the green_mask");
  blue_mask        = ins.read_long("Reading the blue_mask");
  bits_per_rgb     = ins.read_long("Reading the bits_per_rgb");
  colormap_entries = ins.read_long("Reading the colormap_entries");
  ncolors          = ins.read_long("Reading the ncolors");
  window_width     = ins.read_long("Reading the window_width");
  window_height    = ins.read_long("Reading the window_height");
  window_x         = ins.read_long("Reading the window_x");
  window_y         = ins.read_long("Reading the window_y");
  window_bdrwidth  = ins.read_long("Reading the window_bdrwidth");

  					// Read the window name
  register int len = header_size - sizeof(X11WDFileHeader);
  win_name = new char[len];
  if( !ins.read(win_name, len) )
    ins.error("Error reading the window name");

  if( ncolors > 0 )			// Skip the Xcolors
    ins.seekg(ncolors*sizeof(X11XColor),ios::cur);

  depth = bits_per_pixel;	// The depth of the image to construct
  nrows = pixmap_height;
  ncols = pixmap_width;
}

/*
 *------------------------------------------------------------------------
 *	Print out all the control information pertaining to the
 *		 	X Window Image read
 */

void XWDump::info(void) const
{
  cout << "\n\n=====>The following X Window dump has been read";
  cout << "\nVersion:        " << file_version;
  cout << "\nPixmap format:  ";
  switch(pixmap_format)
  {
    case XYBitmap:
         cout << "XYBitmap";
         break;
    case XYPixmap:
         cout << "XYPixmap";
         break;
    case ZPixmap:
         cout << "ZPixmap";
         break;
    default:
	 _error("Unknown pixel format: %d",pixmap_format);
  }
  cout << "\nPixmap depth:   " << pixmap_depth;
  cout << "\nPixmap width:   " << pixmap_width;
  cout << "\nPixmap height:  " << pixmap_height;
  cout << "\nBitmap x offs:  " << xoffset;
  cout << "\nByte order:     " << ( xbyte_order == MSBFirst ?
				    "MSBFirst" : "LSBFirst" );
  cout << "\nBitmap unit:    " << bitmap_unit;
  cout << "\nBitmap bitOrder:" << ( bitmap_bit_order == MSBFirst ?
				  "MSBFirst" : "LSBFirst" );
  cout << "\nScanline pad:   " << bitmap_pad;
  cout << "\nBits per pixel: " << bits_per_pixel;
  cout << "\nBytes per line: " << bytes_per_line;
  cout << "\nVisual class:   ";
  switch(visual_class)
  {
    case StaticGray:
         cout << "StaticGray";
         break;
    case GrayScale:
         cout << "GrayScale";
         break;
    case StaticColor:
         cout << "StaticColor";
         break;
    case PseudoColor:
         cout << "PseudoColor";
         break;
    case TrueColor:
         cout << "TrueColor";
         break;
    case DirectColor:
         cout << "DirectColor";
         break;
    default:
	 _error("Unknown visual class: %d",visual_class);
  }

  cout << "\nZ red mask:     0x" << ::hex << red_mask;
  cout << "\nZ green mask:   0x" << ::hex << green_mask;
  cout << "\nZ blue mask:    0x" << ::hex << blue_mask << ::dec;
  cout << "\nBits per rgb:   " << bits_per_rgb;
  cout << "\nColormap items: " << colormap_entries;
  cout << "\nNo. colors:     " << ncolors;
  cout << "\nWindow width:   " << window_width;
  cout << "\nWindow height:  " << window_height;
  cout << "\nWindow left  X: " << window_x;
  cout << "\nWindow upper Y: " << window_y;
  cout << "\nWindow border:  " << window_bdrwidth;
  cout << "\nWindow name:    " << win_name;

  cout << "\n-----End of X Window dump information" << endl;
}

/*
 *------------------------------------------------------------------------
 *		Read the image pixmap and write it into the image
 */

				// Load image from a Gray scale pixelmap
				// of depth 8 (that is, pixel matrix
				// arranged row-by-row)
class load_8bit_pixmap : public PixelAction
{
  EndianIn& ins;
  void operation(GRAY& pixel, const card, const card)
	{ pixel = ins.read_byte("Reading the pixel map"); }
public:
  load_8bit_pixmap(EndianIn& _ins) : ins(_ins)
  { cerr << "\nReading the gray scale pixel map of depth 8..."; }
};

				// Try to load arbitrary pixel map from the
				// XWD file, doing pixel unpacking (and
				// conversion from rgb to the grayscale)
				// if necessary
class load_any_pixmap : public PixelAction
{
  const int pixel_mask;			// Parameters of the pixel map
  const int bits_per_pixel;		// (cached from XWDump header)
  const int bits_per_item;
  const int bit_order;
  bool convert_rgb;

  EndianIn& ins;
  char * scanline;
  char * scanline_end;
				// Used in pixel unpacking
  char *cp;				// Current scanline ptr
  long int bits_storage;
  int bit_shift;
  int remains_bit;

  void load_scanline(const card row);
  void operation(GRAY& pixel, const card row, const card col);
public:
  load_any_pixmap(const XWDump& header);
  virtual ~load_any_pixmap(void)	{ delete scanline; }
};
				// Prepare for loading pixelmap, initialize
				// param cache and bit cache
load_any_pixmap::load_any_pixmap(const XWDump& header)
   : pixel_mask(( 1 << header.bits_per_pixel ) - 1),
     bits_per_pixel( header.bits_per_pixel ),
     bits_per_item( header.bitmap_unit ),
     bit_order( header.bitmap_bit_order ),
     convert_rgb( header.visual_class == TrueColor ||
		 header.visual_class == DirectColor ),
     ins(header.ins),
     bits_storage(0), bit_shift(0), remains_bit(0)
{
  scanline = new char[header.bytes_per_line];
  scanline_end = scanline + header.bytes_per_line;
  cp = &scanline[0];	

  if( convert_rgb )
    message("\nA colored image with direct RGB specification\n"
	    "is to be converted to gray scale\n");
}

void load_any_pixmap::load_scanline(const card row)
{
  switch( bits_per_item )		// Read a scanline
  {
    case 8:
    {
      for(register char *p = (char *)scanline; p < (char*)scanline_end;)
	*p++ = ins.read_byte("Reading the scanline");
    }
    break;

    case 16:
    {
      for(register short *p = (short *)scanline; p < (short*)scanline_end;)
	  *p++ = ins.read_short("Reading the scanline");
    }
    break;

    case 32:
    {
      for(register long *p = (long *)scanline; p < (long*)scanline_end;)
	*p++ = ins.read_long("Reading the scanline");
    }
    break;

    default:
      assure(0,"can't happen");
  }

  assert( row == 0 || cp == scanline_end );
  cp = &scanline[0];
  bits_storage = 0;
  bit_shift = 0;
  remains_bit = 0;
}


				// Load the current (row,col) pixel
void load_any_pixmap::operation(GRAY& pixel, const card row, const card col)
{
  if( col == 0 )
    load_scanline(row);

  if( remains_bit == 0 )
  {
    switch(bits_per_item)
    {
      case 8:
      bits_storage = *cp;
      break;

      case 16:
      bits_storage = *((short int *)cp);
      break;

      case 32:
      bits_storage = *((long int *)cp);
      break;
    }
    cp += bits_per_item/8;
    remains_bit = bits_per_item;
    bit_shift = ( bit_order == MSBFirst ? bits_per_item - bits_per_pixel :
		 0 );
  }
  assert( bit_shift >= 0 );
  register unsigned int raw_pixel = ( bits_storage >> bit_shift ) & pixel_mask;
  bit_shift += ( bit_order == MSBFirst ? - bits_per_pixel : 
		bits_per_pixel );
  remains_bit -= bits_per_pixel;
  assure(remains_bit >= 0,"Something goes wrong with bits conversion");

  if(convert_rgb)
    pixel = ((raw_pixel & 0xff0000) << (20-16)) +	/* Red part	*/
            ((raw_pixel & 0xff00) << (10-8)) +		/* Green part	*/
	    ((raw_pixel & 0xff));			/* Blue part	*/
  else
    pixel = raw_pixel;
}


				// Analyze the header and figure out if
				// we can handle the XWdump we've read.
				// Then read the pixelmap using
				// the best strategy
void XWDump::fill_in(IMAGE& im) const
{ 
  const int bits_per_item = bitmap_unit;

  if( pixmap_depth > 24 )
    _error("can't handle X11 pixmap_depth %d greater than 24",pixmap_depth);
  if( bits_per_rgb > 24 )
    _error("can't handle X11 bits_per_rgb > 24");
  if( pixmap_format != ZPixmap && pixmap_depth != 1 )
    _error("can't handle X11 pixmap_format %d with depth != 1",pixmap_format);
  if( bits_per_item != 8 && bits_per_item != 16 && bits_per_item != 32 )
    _error("X11 bitmap_unit (%d) is non-standard - can't handle",
	   bits_per_item);
  if( bits_per_item == 8 && bits_per_pixel == 8 && 
      visual_class != TrueColor && visual_class != DirectColor )
    im.apply(load_8bit_pixmap(ins));
#if 0
  else if( bits_per_item == 8 && bits_per_pixel == 1 )
    pixmap_type = PT_BlackWhite;
#endif
  else
    im.apply(load_any_pixmap(*this));

    im.set_name(win_name);

  cerr << "\n" << im.q_ncols() << "x" << im.q_nrows() << "x"
       << im.q_depth() << " image '"
       << im.q_name() << "' has been read" << endl;
}

/*
 *========================================================================
 *		     Root module - actual IMAGE constructor
 */

const LazyImagePtr
read_xwd(const char file_name[], const bool print_header_info)
{
  cerr << "Reading an X window dump image" << endl;

  XWDump * xwdump = new XWDump(file_name);

  if( print_header_info )
    xwdump->info();

  return xwdump;
}

